@@ -5,9 +5,10 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 0.7
-;; Package-Requires: ((emacs "27.1") (compat "29.1.3.4"))
+;; Version: 0.8
+;; Package-Requires: ((emacs "27.1") (compat "29.1.4.0"))
 ;; Homepage: https://github.com/minad/tempel
+;; Keywords: abbrev, languages, tools, wp
 
 ;; This file is part of GNU Emacs.
 
@@ -75,11 +76,11 @@ trigger completion."
 
 (defcustom tempel-insert-annotation 40
   "Annotation width for `tempel-insert'."
-  :type '(choice (const nil integer)))
+  :type '(choice (const nil natnum)))
 
 (defcustom tempel-complete-annotation 20
   "Annotation width for `tempel-complete'."
-  :type '(choice (const nil integer)))
+  :type '(choice (const nil natnum)))
 
 (defcustom tempel-user-elements nil
   "List of user element handler functions.
@@ -215,30 +216,34 @@ REGION are the current region bounds."
 (defun tempel--range-modified (ov &rest _)
   "Range overlay OV modified."
   (when (and (not tempel--inhibit-hooks) (= (overlay-start ov) (overlay-end ov)))
-    (tempel--disable (overlay-get ov 'tempel--range))))
+    (let ((inhibit-modification-hooks nil)
+          (tempel--inhibit-hooks t))
+      (tempel--disable (overlay-get ov 'tempel--range)))))
 
 (defun tempel--field-modified (ov after beg end &optional _len)
   "Update field overlay OV.
 AFTER is non-nil after the modification.
 BEG and END are the boundaries of the modification."
   (unless tempel--inhibit-hooks
-    (cond
-     ;; Erase default before modification if at beginning or end
-     ((and (not after) (overlay-get ov 'tempel--default)
-           (or (= beg (overlay-start ov)) (= end (overlay-end ov))))
-      (delete-region (overlay-start ov) (overlay-end ov)))
-     ;; Update field after modification
-     (after
-      (let ((st (overlay-get ov 'tempel--field)))
-        (unless undo-in-progress
-          (move-overlay ov (overlay-start ov) (max end (overlay-end ov))))
-        (when-let (name (overlay-get ov 'tempel--name))
-          (setf (alist-get name (cdr st))
-                (buffer-substring-no-properties
-                 (overlay-start ov) (overlay-end ov))))
-        (unless undo-in-progress
-          (tempel--synchronize-fields st ov)))))
-    (tempel--update-mark ov)))
+    (let ((inhibit-modification-hooks nil)
+          (tempel--inhibit-hooks t))
+      (cond
+       ;; Erase default before modification if at beginning or end
+       ((and (not after) (overlay-get ov 'tempel--default)
+             (or (= beg (overlay-start ov)) (= end (overlay-end ov))))
+        (delete-region (overlay-start ov) (overlay-end ov)))
+       ;; Update field after modification
+       (after
+        (let ((st (overlay-get ov 'tempel--field)))
+          (unless undo-in-progress
+            (move-overlay ov (overlay-start ov) (max end (overlay-end ov))))
+          (when-let ((name (overlay-get ov 'tempel--name)))
+            (setf (alist-get name (cdr st))
+                  (buffer-substring-no-properties
+                   (overlay-start ov) (overlay-end ov))))
+          (unless undo-in-progress
+            (tempel--synchronize-fields st ov)))))
+      (tempel--update-mark ov))))
 
 (defun tempel--synchronize-fields (st current)
   "Synchronize fields of ST, except CURRENT overlay."
@@ -250,22 +255,22 @@ BEG and END are the boundaries of the modification."
           (let (x)
             (setq x (or (and (setq x (overlay-get ov 'tempel--form)) (eval x (cdr st)))
                         (and (setq x (overlay-get ov 'tempel--name)) (alist-get x (cdr st)))))
-            (when x (tempel--replace (overlay-start ov) (overlay-end ov) ov x)))))
+            (when x (tempel--synchronize-replace (overlay-start ov) (overlay-end ov) ov x)))))
       ;; Move range overlay
       (move-overlay range (overlay-start range)
                     (max (overlay-end range) (overlay-end ov))))))
 
-(defun tempel--replace (beg end ov str)
-  "Replace region beween BEG and END with STR.
+(defun tempel--synchronize-replace (beg end ov str)
+  "Replace region between BEG and END with STR.
 If OV is alive, move it."
   (let ((old (buffer-substring-no-properties beg end)))
     (setq ov (and ov (overlay-buffer ov) ov))
     (unless (equal str old)
       (unless (eq buffer-undo-list t)
-        (push (list 'apply #'tempel--replace beg (+ beg (length str)) ov old)
+        (push (list 'apply #'tempel--synchronize-replace
+                    beg (+ beg (length str)) ov old)
               buffer-undo-list))
-      (let ((buffer-undo-list t)
-            (tempel--inhibit-hooks t))
+      (let ((buffer-undo-list t))
         (save-excursion
           (goto-char beg)
           (delete-char (- end beg))
@@ -289,7 +294,8 @@ If OV is alive, move it."
 NAME is the optional field name.
 INIT is the optional initial input.
 Return the added field."
-  (let ((ov (make-overlay (point) (point))))
+  (let ((ov (make-overlay (point) (point)))
+        (hooks (list #'tempel--field-modified)))
     (push ov (car st))
     (when name
       (overlay-put ov 'tempel--name name)
@@ -300,9 +306,9 @@ Return the added field."
       (move-overlay ov (overlay-start ov) (point)))
     (tempel--update-mark ov)
     (overlay-put ov 'tempel--field st)
-    (overlay-put ov 'modification-hooks (list #'tempel--field-modified))
-    (overlay-put ov 'insert-in-front-hooks (list #'tempel--field-modified))
-    (overlay-put ov 'insert-behind-hooks (list #'tempel--field-modified))
+    (overlay-put ov 'modification-hooks hooks)
+    (overlay-put ov 'insert-in-front-hooks hooks)
+    (overlay-put ov 'insert-behind-hooks hooks)
     (overlay-put ov 'face 'tempel-field)
     (when (and init (get-text-property 0 'tempel--default init))
       (overlay-put ov 'face 'tempel-default)
@@ -350,15 +356,15 @@ Return the added field."
     ((or 'p `(,(or 'p 'P) . ,rest)) (apply #'tempel--placeholder st rest))
     ((or 'r 'r> `(,(or 'r 'r>) . ,rest))
      (if (not region)
-         (when-let (ov (apply #'tempel--placeholder st rest))
-           (unless rest
-             (overlay-put ov 'tempel--quit t)))
+         (when-let ((ov (apply #'tempel--placeholder st rest))
+                    ((not rest)))
+           (overlay-put ov 'tempel--enter #'tempel--done))
        (goto-char (cdr region))
        (when (eq (or (car-safe elt) elt) 'r>)
          (indent-region (car region) (cdr region) nil))))
     ;; TEMPEL EXTENSION: Quit template immediately
-    ('q (overlay-put (tempel--field st) 'tempel--quit t))
-    (_ (if-let (ret (run-hook-with-args-until-success 'tempel-user-elements elt))
+    ('q (overlay-put (tempel--field st) 'tempel--enter #'tempel--done))
+    (_ (if-let ((ret (run-hook-with-args-until-success 'tempel-user-elements elt)))
            (tempel--element st region ret)
          ;; TEMPEL EXTENSION: Evaluate forms
          (tempel--form st elt)))))
@@ -397,16 +403,16 @@ If a field was added, return it."
             (setf (overlay-end ov) (point)))))
       ;; Activate template
       (let ((st (cons nil nil))
-            (range (point))
+            (ov (point))
             (tempel--inhibit-hooks t))
         (while (and template (not (keywordp (car template))))
           (tempel--element st region (pop template)))
-        (setq range (make-overlay range (point) nil t))
-        (push range (car st))
-        (overlay-put range 'modification-hooks (list #'tempel--range-modified))
-        (overlay-put range 'tempel--range st)
-        (overlay-put range 'tempel--post (plist-get plist :post))
-        ;;(overlay-put range 'face 'region) ;; TODO debug
+        (setq ov (make-overlay ov (point) nil t))
+        (push ov (car st))
+        (overlay-put ov 'modification-hooks (list #'tempel--range-modified))
+        (overlay-put ov 'tempel--range st)
+        (overlay-put ov 'tempel--post (plist-get plist :post))
+        ;;(overlay-put ov 'face 'region) ;; TODO debug
         (push st tempel--active)))
     (cond
      ((cl-loop for ov in (caar tempel--active)
@@ -423,7 +429,7 @@ If a field was added, return it."
   (cl-loop
    with all = nil
    for (file . _ts) in (car tempel--path-templates) do
-   (when-let (buf (get-file-buffer file))
+   (when-let ((buf (get-file-buffer file)))
      (with-current-buffer buf
        (when (and (buffer-modified-p)
                   (pcase (or all (read-answer
@@ -531,13 +537,13 @@ This is meant to be a source in `tempel-template-sources'."
 (defun tempel-beginning ()
   "Move to beginning of the template."
   (interactive)
-  (when-let (pos (tempel--beginning))
+  (when-let ((pos (tempel--beginning)))
     (if (= pos (point)) (tempel-done) (goto-char pos))))
 
 (defun tempel-end ()
   "Move to end of the template."
   (interactive)
-  (when-let (pos (tempel--end))
+  (when-let ((pos (tempel--end)))
     (if (= pos (point)) (tempel-done) (goto-char pos))))
 
 (defun tempel--field-at-point ()
@@ -552,7 +558,7 @@ This is meant to be a source in `tempel-template-sources'."
 (defun tempel-kill ()
   "Kill the field contents."
   (interactive)
-  (if-let (ov (tempel--field-at-point))
+  (if-let ((ov (tempel--field-at-point)))
       (kill-region (overlay-start ov) (overlay-end ov))
     (kill-sentence nil)))
 
@@ -560,14 +566,14 @@ This is meant to be a source in `tempel-template-sources'."
   "Move ARG fields forward and quit at the end."
   (interactive "p")
   (cl-loop for i below (abs arg) do
-           (if-let (next (tempel--find arg)) (goto-char next)
+           (if-let ((next (tempel--find arg)))
+               (goto-char next)
              (tempel-done)
              (cl-return)))
-  ;; If the current field is marked as "quitting", disable its
-  ;; containing template right away.
+  ;; Run the enter action of the field.
   (when-let ((ov (tempel--field-at-point))
-             ((overlay-get ov 'tempel--quit)))
-    (tempel--done (overlay-get ov 'tempel--field))))
+             (fun (overlay-get ov 'tempel--enter)))
+    (funcall fun ov)))
 
 (defun tempel-previous (arg)
   "Move ARG fields backward and quit at the beginning."
@@ -587,11 +593,12 @@ This is meant to be a source in `tempel-template-sources'."
 (defun tempel-abort ()
   "Abort template insertion."
   (interactive)
-  (when-let ((beg (tempel--beginning))
-             (end (tempel--end)))
-    ;; TODO abort only the topmost template?
-    (while tempel--active (tempel--disable))
-    (delete-region beg end)))
+  ;; TODO abort only the topmost template?
+  (while-let ((st (car tempel--active)))
+    (let ((beg (overlay-start (caar st)))
+          (end (overlay-end (caar st))))
+      (tempel--disable)
+      (delete-region beg end))))
 
 (defun tempel--disable (&optional st)
   "Disable template ST, or last template."
@@ -611,9 +618,9 @@ This is meant to be a source in `tempel-template-sources'."
   ;; TODO disable only the topmost template?
   (while tempel--active (tempel--done)))
 
-(defun tempel--done (&optional st)
-  "Finalize template ST, or last template."
-  (let ((st (or st (car tempel--active)))
+(defun tempel--done (&optional ov)
+  "Finalize template associated with field OV, or last template."
+  (let ((st (if ov (overlay-get ov 'tempel--field) (car tempel--active)))
         (buf (current-buffer)))
     ;; Ignore errors in post expansion to ensure that templates can be
     ;; terminated gracefully.
@@ -651,7 +658,13 @@ The completion table specifies the category `tempel'."
 ;;;###autoload
 (defun tempel-expand (&optional interactive)
   "Expand exactly matching template name at point.
-If INTERACTIVE is nil the function acts like a capf."
+This completion-at-point-function (Capf) returns only the single
+exactly matching template name.  As a consequence the completion
+UI (e.g. Corfu) does not present the candidates for selection.
+If you want to select from a list of templates, use
+`tempel-complete' instead.  If INTERACTIVE is nil the function
+acts like a Capf, otherwise like an interactive completion
+command."
   (interactive (list t))
   (if interactive
       (tempel--interactive #'tempel-expand)
@@ -670,7 +683,11 @@ If INTERACTIVE is nil the function acts like a capf."
 ;;;###autoload
 (defun tempel-complete (&optional interactive)
   "Complete template name at point and expand.
-If INTERACTIVE is nil the function acts like a capf."
+This completion-at-point-function (Capf) returns a list of all
+possible template names, which are then displayed in the
+completion UI (e.g. Corfu) for selection.  See also
+`tempel-expand'.  If INTERACTIVE is nil the function acts like a
+Capf, otherwise like an interactive completion command."
   (interactive (list t))
   (if interactive
       (progn
